(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory ErrorMonad imports Util begin

types ('e,'a) emonad = "('e,'a) either" (* error and value *)
types 'e monad = "('e,unit) either"     (* just error *)

definition return :: "'a \<Rightarrow> ('e,'a)emonad" where "return \<equiv> Right"
abbreviation OK :: "'e monad" where "OK \<equiv> return ()"
definition error :: "'e \<Rightarrow> ('e,'a)emonad" where "error \<equiv> Left"
definition isValue :: "('e,'a)emonad \<Rightarrow> bool" where "isValue p \<equiv> case p of Left l \<Rightarrow> False | Right r \<Rightarrow> True"
abbreviation isOK where "isOK \<equiv> isValue"


lemma [simp] : "(return r = error x) = False" by (simp add: return_def error_def)
lemma [simp] : "(error x = return r) = False" by (simp add: return_def error_def)
lemma [simp] : "isOK (return x)" by (simp add: return_def error_def isValue_def)
lemma [simp] : "\<not> isOK (error x)" by (simp add: return_def error_def isValue_def)


definition checkB :: "bool \<Rightarrow> 'e \<Rightarrow> 'e monad"
 where "checkB b e \<equiv> if b then OK else error e"
lemma checkBSound[simp] : "isOK(checkB b f) = b"
  unfolding checkB_def  by (cases b, auto simp: return_def error_def isValue_def)


definition checkF :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'a \<Rightarrow> 'b monad"
  where "checkF f g x  \<equiv> if f x then OK else error (g x)"

abbreviation checkS :: "('a \<Rightarrow> bool) \<Rightarrow> 'b \<Rightarrow> 'a \<Rightarrow> 'b monad"
  where "checkS f err \<equiv> checkF f (\<lambda>x. err)"
abbreviation check :: "('a \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> 'a monad"
  where "check f \<equiv> checkF f id"

lemma checkFSound[simp] : "isOK (checkF f s x) = f x" unfolding checkF_def  by (cases "f x", auto simp: return_def error_def isValue_def)

definition bind :: "('e,'a)emonad \<Rightarrow> ('a \<Rightarrow> ('e,'b)emonad) \<Rightarrow> ('e,'b)emonad" (infixl "\<guillemotright>=" 25)
  where "p \<guillemotright>= vq \<equiv> case p of Right val \<Rightarrow> vq val | Left e \<Rightarrow> Left e"

lemma monad1[simp] : "(return x \<guillemotright>= iq) = iq x" by (auto simp: return_def error_def isValue_def bind_def)
lemma monad2[simp] : "(q \<guillemotright>= return ) = q" by (cases q, auto simp: return_def error_def isValue_def bind_def)
lemma monad3[simp] : "((q \<guillemotright>= ip) \<guillemotright>= jr) = (q \<guillemotright>= (\<lambda> x . ip x \<guillemotright>= jr))" by (cases q, auto simp: return_def error_def isValue_def bind_def)
lemma bindSound: "isOK(p \<guillemotright>= qx) = (isOK(p) \<and> (\<forall> x. ((p = return x) \<longrightarrow> isOK(qx x))))" by (cases p, auto simp: return_def error_def isValue_def  bind_def)
lemma bindSound2: "(error y \<guillemotright>= q) = error y" by (auto simp: return_def error_def isValue_def bind_def)
abbreviation  "then" :: "('e,'a)emonad \<Rightarrow> ('e,'b)emonad \<Rightarrow> ('e,'b)emonad" (infixl "\<guillemotright>" 24)
  where "p \<guillemotright> q \<equiv> p \<guillemotright>= (\<lambda> x. q)"
lemma thenSound [simp]: "isOK(p \<guillemotright> q) = (isOK(p) \<and> isOK(q))" by (cases p, auto simp: return_def error_def isValue_def  bind_def)
abbreviation changeValue :: "('e,'a)emonad \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('e,'b)emonad" 
  where "changeValue p f \<equiv> p \<guillemotright>= (\<lambda> y. return (f y))" 

notation (xsymbols)
  changeValue  (infixl "\<rightarrow>" 28) 
  

lemma changeValueSound[simp] : "isOK(p \<rightarrow> f) = isOK(p)" by (cases p, auto simp: return_def error_def isValue_def  bind_def)


definition catch :: "('e,'a)emonad \<Rightarrow> ('e \<Rightarrow> ('f,'a)emonad) \<Rightarrow> ('f,'a)emonad" (infixl "\<guillemotleft>=" 23)
  where "p \<guillemotleft>= eq \<equiv> case p of Right val \<Rightarrow> Right val | Left err \<Rightarrow> eq err"
abbreviation catchThen :: "('e,'a)emonad \<Rightarrow> ('f,'a)emonad \<Rightarrow> ('f,'a)emonad" (infixl "\<guillemotleft>" 22)
  where "p \<guillemotleft> q \<equiv> p \<guillemotleft>= (\<lambda> x. q)"
abbreviation changeError :: "('e,'a)emonad \<Rightarrow> ('e \<Rightarrow> 'f) \<Rightarrow> ('f,'a)emonad" (infixl "\<leftarrow>" 27)
  where "p \<leftarrow> f \<equiv> p \<guillemotleft>= (\<lambda> x. error (f x))"


lemma catchSound1 [simp]: "(return x \<guillemotleft>= q) = return x" by (auto simp: return_def error_def isValue_def catch_def)
lemma catchSound2 [simp]: "(error x \<guillemotleft>= q) = q x" by (auto simp: return_def error_def isValue_def catch_def)
lemma catchSound3 [simp]: "isOK(q \<leftarrow> f) = isOK(q)" by (cases q, auto simp: return_def error_def isValue_def catch_def)
lemma catchSound4 [simp]: "isOK(p \<guillemotleft> q) = isOK(p) \<or> isOK(q)" by (cases q, cases p, auto simp: return_def error_def isValue_def catch_def)

(*
definition checkNegate :: "'e monad \<Rightarrow> unit monad" ("neg")
  where "neg p \<equiv> case p of None \<Rightarrow> Error () | Some x \<Rightarrow> OK"


lemma checkNegateSound[simp] : "isOK (neg p) = isnotOK p" 
  unfolding checkNegate_def OK_def Error_def isOK_def isOK_def by (cases "p", auto)
*)

fun getValue :: "('e,'a)emonad \<Rightarrow> 'a" where "getValue (Right r) = r"
lemma getValue1[simp]: "isOK(p) \<Longrightarrow> getValue(p \<guillemotright> q) = getValue q" unfolding isValue_def bind_def by (cases p, auto)
lemma getValue2[simp]: "getValue (return x) = x" unfolding isValue_def return_def by simp
lemma getValue3[simp]: "isOK(p) \<Longrightarrow> getValue(p \<leftarrow> f) = getValue p" unfolding catch_def isValue_def bind_def error_def by (cases p, auto)




definition checkAllMI :: "('a \<Rightarrow> 'e monad) \<Rightarrow> 'a list \<Rightarrow> ('a \<times> 'e) monad"
 where "checkAllMI f xs = foldr (\<lambda> x iq. f x \<leftarrow> Pair x \<guillemotright> iq) xs OK"


lemma checkAllMISound[simp] : "isOK (checkAllMI f xs) = (\<forall> x \<in> set xs. isOK (f x))"
unfolding checkAllMI_def by (induct xs, auto)


abbreviation checkAllM :: "('a \<Rightarrow> 'e monad) \<Rightarrow> 'a list \<Rightarrow> 'e monad"
 where "checkAllM f xs \<equiv> checkAllMI f xs \<leftarrow> snd"

abbreviation checkAll :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a monad"
 where "checkAll f \<equiv> checkAllM (check f)"

definition extractOneMI :: "('a \<Rightarrow> ('e,'v)emonad) \<Rightarrow> 'a list \<Rightarrow> ('e list,('a \<times> 'v))emonad"
 where "extractOneMI f xs \<equiv> foldr (\<lambda> x iq. 
          (f x \<rightarrow> Pair x \<guillemotleft>= (\<lambda> e.  iq \<leftarrow> (\<lambda> es. e # es)))) xs (error [])"

(*
lemma extractOneMISound[simp] : "isOK (checkOneMI aq xs) = (\<exists> x \<in> set xs. isOK (aq x))"
  unfolding checkOneMI_def isOK_def Error_def OK_def bind_def mapM_def return_def proof (induct xs, auto)
*)

abbreviation extractOneM :: "('a \<Rightarrow> ('e,'v)emonad) \<Rightarrow> 'a list \<Rightarrow> (unit, 'a)emonad"
  where "extractOneM f xs \<equiv> extractOneMI f xs \<rightarrow> fst \<leftarrow> (\<lambda>x. ()) "

abbreviation extractOne :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> (unit,'a)emonad"
  where "extractOne f \<equiv> extractOneM (check f)"


definition checkSubset :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a monad"
  where "checkSubset xs ys \<equiv> checkAll (\<lambda> x. x mem ys) xs"

lemma checkSubsetSound[simp] : "isOK(checkSubset xs ys) = (set xs \<subseteq> set ys)"
  unfolding checkSubset_def by (auto simp: mem_iff)

definition checkSameSet :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a monad"
  where "checkSameSet xs ys \<equiv> checkSubset xs ys \<guillemotright> checkSubset ys xs"
lemma checkSameSetSound[simp] : "isOK(checkSameSet xs ys) = (set xs = set ys)"
  unfolding checkSameSet_def by (auto)


definition checkDisjoint :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a monad"
where "checkDisjoint xs ys \<equiv> checkAllM  (check (\<lambda> x. \<not> x mem ys)) xs"
      

lemma checkDisjointSound[simp] :
  "isOK (checkDisjoint xs ys) = (set xs \<inter> set ys = {})"
unfolding checkDisjoint_def by (auto simp: mem_iff)



fun foldR :: "('a \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b"
where "foldR f [] b = b"
    | "foldR f (x # xs) b = (f x xs (foldR f xs b))" 


lemma thenCong [fundef_cong] : fixes m1 m2 n1 n2 :: "('e,'a)emonad"
  assumes "m1 = m2" and "n1 = n2"
  shows "(m1 \<guillemotright> n1) = (m2 \<guillemotright> n2)"
using assms
 by (unfold "bind_def", auto)

lemma changeErrorCong [fundef_cong] : 
  assumes "c = d" 
  shows "(c \<leftarrow> s) = (d \<leftarrow> s)"
using assms
unfolding catch_def return_def bind_def by (cases c, auto)

lemma changeValueCong [fundef_cong] : 
  assumes "c = d" 
  shows "(c \<rightarrow> s) = (d \<rightarrow> s)"
using assms
unfolding catch_def return_def bind_def by (cases c, auto)


(*
definition checkAllMI :: "('a \<Rightarrow> 'e monad) \<Rightarrow> 'a list \<Rightarrow> ('a \<times> 'e) monad"
*)

lemma checkAllMI[fundef_cong] : fixes c :: "'a \<Rightarrow> 'e monad"
  assumes cond: "\<And> x. x \<in> set xs \<Longrightarrow> c x = d x"
  shows "checkAllMI c xs = checkAllMI d xs"
unfolding checkAllMI_def
proof (rule foldr_cong, auto)
  fix a :: "('a \<times> 'e) monad" and x
  assume "x \<in> set xs"
  hence c: "c x = d x" and a: "a = a" using cond by auto
  from c have "(c x \<leftarrow> Pair x) = (d x \<leftarrow> Pair x)" by (rule changeErrorCong)
  from this and  a show "(c x \<leftarrow> Pair x \<guillemotright> a) = (d x \<leftarrow> Pair x \<guillemotright> a)"
    by (rule  thenCong)
qed 

definition checkAllCombinations :: "('a \<Rightarrow> 'a \<Rightarrow> 'b monad) \<Rightarrow> 'a list \<Rightarrow> (('a \<times> 'a) \<times> 'b) monad"
where "checkAllCombinations c xs \<equiv> foldR (\<lambda> x ys res. (checkAllM (\<lambda> y. c x y \<leftarrow> (\<lambda> b. ((x,y), b)) \<guillemotright> res) ys)) xs OK"

lemma checkAllCombinationsSound[simp]:
  "isOK(checkAllCombinations c xs) = (\<forall>j<length xs.\<forall>i<j. isOK(c (xs!i) (xs!j)))"
unfolding checkAllCombinations_def
proof (rule iffI)
  let ?f = "\<lambda>x ys res. checkAllM (\<lambda> y. c x y \<leftarrow> (\<lambda> b. ((x,y), b)) \<guillemotright> res) ys"
  assume "isOK(foldR ?f xs OK)"
  thus "\<forall>j<length xs. \<forall>i<j. isOK (c (xs ! i) (xs ! j))"
  proof(induct xs, simp)
  case (Cons a xs)
   assume assFold:"isOK(foldR ?f (a # xs) OK)"
   show "\<forall>j<length (a # xs). \<forall>i<j. isOK(c ((a # xs) ! i) ((a # xs) ! j))"
   proof(rule allI)
    fix j
    show "j < length (a # xs) \<longrightarrow> (\<forall>i<j. isOK(c ((a # xs) ! i) ((a # xs) ! j)))"
    proof(rule impI, cases j, simp)
     case (Suc k)
      assume "j < length (a # xs)" 
      with Suc have "j = Suc k" and "k < length xs" by auto
      let ?f' = "\<lambda>y. c a y \<leftarrow> (\<lambda> b. ((a,y), b)) \<guillemotright> (foldR (\<lambda>x ys res. checkAllM (\<lambda> y. c x y \<leftarrow> (\<lambda> b. ((x,y), b)) \<guillemotright> res) ys) xs OK)"
      from assFold have f'none:"\<forall> x \<in> set xs. isOK (?f' x)" by simp
      show "\<forall>i<j. isOK (c ((a # xs) ! i) ((a # xs) ! j))"
      proof
       fix i
       show "i < j \<longrightarrow> isOK ( c ((a # xs) ! i) ((a # xs) ! j))"
       proof(rule impI, cases i, simp)
        case 0
        from f'none have "\<forall>x\<in>set xs. isOK (c a x)" by simp  
        with 0  `k < length xs` Suc show "isOK (c a ((a # xs) ! j))" by simp
       next
        case (Suc m)
        assume "i < j"
        with Suc `j = Suc k` `j < length (a # xs)` have "m < k" and "length xs > 0" by auto
        hence "hd xs \<in> set xs" by auto
        with f'none have "isOK( foldR ?f xs OK)" by simp
        with Cons have allj:"\<forall>j<length xs. \<forall>i<j. isOK(c (xs ! i) (xs ! j))" by simp
        from `j < length (a # xs)` `j = Suc k` have "k < length xs" by auto
        with allj `m < k` Suc `j = Suc k` show "isOK (c ((a # xs) ! i) ((a # xs) ! j))" by auto
       qed
      qed
     qed
   qed
 qed
 next
  let ?f = "\<lambda>x ys res. checkAllM (\<lambda> y. c x y \<leftarrow> (\<lambda> b. ((x,y), b)) \<guillemotright> res) ys"
  assume allj:"\<forall>j<length xs. \<forall>i<j. isOK (c (xs ! i) (xs ! j))"
  thus "isOK (foldR ?f xs OK)"
  proof(induct xs, simp)
   case (Cons a as)
    with allj have allas:"\<forall>j<length (a # as). \<forall>i<j. isOK (c ((a # as) ! i) ((a # as) ! j))" by auto
    have checkFirst:"\<forall>x\<in>set as. isOK (c a x)" 
     proof
      fix x assume "x \<in> set as"
      hence "\<exists> j < length as. as ! j = x" using in_set_conv_nth[where ?xs = "as"] by auto 
      then obtain j where "j < length as" and "as ! j = x" by auto
      with allas have "\<forall>i<(Suc j). isOK (c ((a # as) ! i) ((a # as) ! (Suc j)))" by auto
      with `as ! j = x` show "isOK (c a x)" by auto   
     qed
   let ?f' = "\<lambda>y. c a y \<leftarrow> (\<lambda> b. ((a,y), b)) \<guillemotright> (foldR (\<lambda>x ys res. checkAllM (\<lambda> y. c x y \<leftarrow> (\<lambda> b. ((x,y), b)) \<guillemotright> res) ys) xs OK)" 
   have "\<forall>j<length as. \<forall>i<j. isOK(c (as ! i) (as ! j))"
    proof(rule allI, rule impI)
     fix j assume "j < length as"
     show "\<forall>i<j. isOK (c (as ! i) (as ! j))"
     proof(rule allI, rule impI)
      fix i assume "i < j"
      from `j < length as` allas have "\<forall>i<(Suc j). isOK (c ((a # as) ! i) ((a # as) ! (Suc j)))" by auto
      with `i < j` have  "isOK (c ((a # as) ! Suc i) ((a # as) ! Suc j))" by auto
      thus "isOK (c (as ! i) (as ! j))" by simp
     qed
    qed
    with Cons have "isOK(foldR ?f as OK)" by simp
    with checkFirst show "isOK (foldR ?f (a # as) OK)" by simp
  qed
qed

definition checkPartition :: "'a list list \<Rightarrow> 'a monad"
where "checkPartition cs \<equiv>  
    checkAllCombinations checkDisjoint cs \<leftarrow> snd"

lemma checkPartitionSound[simp]:
  assumes "isOK(checkPartition cs)"
  shows "is_partition(map set cs)"
using assms unfolding is_partition_def checkPartition_def
by (auto simp: set_conv_nth)

definition checkAllMIndex :: "(nat \<Rightarrow> 'a \<Rightarrow> nat) \<Rightarrow> (nat \<Rightarrow> 'a \<Rightarrow> 'e monad) \<Rightarrow> 'a list \<Rightarrow> 'e monad"
 where "checkAllMIndex g f xs \<equiv> snd (foldl (\<lambda> (n,mon) x . (g n x, mon \<guillemotright> f n x)) (1,OK) xs)"

lemma foldlPropagatesError : "snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (n, error e) xs) = error e"
  by (induct xs arbitrary: n, auto simp: bindSound2)
  

lemma checkAllMIndexSound[simp] : assumes "isOK (checkAllMIndex g f xs)"
  shows "\<forall> x \<in> set xs. \<exists> n. isOK (f n x)"
proof -
  from assms have ok: "checkAllMIndex g f xs = OK" unfolding isValue_def return_def by (cases "checkAllMIndex g f xs", auto)
  {
    fix n       
    assume "snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (n, OK) xs) = OK"
    hence "\<forall> x \<in> set xs. \<exists> n. isOK (f n x)"
    proof (induct xs arbitrary: n, simp)
      case (Cons x xs)
      show ?case
      proof (cases "f n x = OK")
	case True
	hence "isOK (f n x)" by auto 
	hence "\<exists> n. isOK (f n x)" by auto
	with True Cons show ?thesis by auto 
      next
	case False
	hence "\<exists> e. f n x = error e" by (cases "f n x", auto simp: return_def error_def)
	from this obtain e where "f n x = error e" ..
	hence "snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (n, OK) (x # xs)) = 
	  snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (g n x, error e) xs)" 
	  by (auto simp: bindSound2)
	also have "\<dots> = error e" using foldlPropagatesError[where g = g and e = e] by auto
	finally have False using Cons by auto
	thus ?thesis ..
      qed
    qed
  }
  
  thus ?thesis using ok  unfolding checkAllMIndex_def by auto
qed

lemma checkAllMIndexComplete : assumes "\<not> isOK(checkAllMIndex g f xs)"
  shows "\<exists> x \<in> set xs. \<exists> n. \<not> isOK (f n x)"
proof -
  from assms obtain e where error: "checkAllMIndex g f xs = error e" 
    unfolding isValue_def error_def by (cases "checkAllMIndex g f xs", auto)
  {
    fix n e      
    assume "snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (n, OK) xs) = error e"
    hence "\<exists> x \<in> set xs. \<exists> n. f n x = error e"
    proof (induct xs arbitrary: n, simp)
      case (Cons x xs)
      show ?case
      proof (cases "f n x = OK")
	case True
	hence "isOK(f n x)" by auto
	with Cons True show ?thesis by auto
      next
	case False
	hence "\<exists> e. f n x = error e" by (cases "f n x", auto simp: return_def error_def)
	from this obtain e1 where err: "f n x = error e1" ..
	hence "snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (n, OK) (x # xs)) = 
	  snd (foldl (\<lambda>(n, mon) x. (g n x, mon \<guillemotright> f n x)) (g n x, error e1) xs)" 
	  by (auto simp: bindSound2)
	also have "\<dots> = error e1" by (rule foldlPropagatesError)
	finally show ?thesis using Cons err by force
      qed
    qed
  }
  with error have nearly: "\<exists> x \<in> set xs. \<exists> n. f n x = error e" unfolding checkAllMIndex_def by blast
  from this obtain x n where "x \<in> set xs \<and> f n x = error e" by blast
  hence "x \<in> set xs \<and> \<not> isOK(f n x)" by auto
  thus ?thesis by auto 
qed

lemma checkAllMIndex[fundef_cong] : fixes f :: "nat \<Rightarrow> 'a \<Rightarrow> 'e monad"
  assumes cond: "\<And> n x. x \<in> set xs \<Longrightarrow> g n x = g' n x"
  assumes cond2: "\<And> n x. x \<in> set xs \<Longrightarrow> f n x = f' n x"
  shows "checkAllMIndex g f xs = checkAllMIndex g' f' xs"
proof -
  { 
    fix n mon
    from assms 
    have "foldl (\<lambda> (n,mon) x . (g n x, mon \<guillemotright> f n x)) (n,mon) xs
    = foldl (\<lambda> (n,mon) x . (g' n x, mon \<guillemotright> f' n x)) (n,mon) xs"
    proof (induct xs arbitrary: n mon, simp)
      case (Cons x xs)
      thus ?case by auto
    qed
  }
  thus ?thesis unfolding checkAllMIndex_def by auto
qed

end
